/*
 * bnpairings/src/ate.c version 20080222
 * Peter Schwabe
 * Public domain
*/

#include <stdio.h>
#include <gmp.h>
#include <cpucycles.h>

#include "fp2e.h"
#include "fp6e.h"
#include "fp12e.h"
#include "curvepoint_fp.h"
#include "twistpoint_fp2.h"
#include "init.h"
#include "points.h"
#include "final_expo.h"
#include "ate.h"


static void linefunction_evaluate_ate(
		fp12e_t rop, 
		const fp2e_t slope_numerator,
		const fp2e_t slope_denominator,
		const fp2e_t zapow3,
		const twistpoint_fp2_t op1,
		const curvepoint_fp_t op2)
{
	fp2e_t tmp1, tmp2, tmp3, tmp4; // Temporary variables needed for intermediary results
	fp6e_t tfp6e1, tfp6e2; 

	fp2e_mul(tmp1, zapow3, slope_denominator);

	fp2e_mul_fpe(tmp1, tmp1, op2->m_y);
	fp2e_neg(tmp1, tmp1);

	fp2e_mul(tmp2, zapow3, slope_numerator);
	fp2e_mul_fpe(tmp2, tmp2, op2->m_x);

	fp2e_mul(tmp3, op1->m_y, slope_denominator);
	fp2e_mul(tmp4, op1->m_x, op1->m_z);
	fp2e_mul(tmp4, tmp4, slope_numerator);
	fp2e_sub(tmp3, tmp3, tmp4);

	fp2e_setzero(tmp4);

	fp6e_set_fp2e(tfp6e1, tmp4, tmp3, tmp2);
	fp6e_set_fp2e(tfp6e2, tmp4, tmp4, tmp1);

	fp12e_set_fp6e(rop, tfp6e1, tfp6e2);
}


static void linefunction_add_ate(
		fp12e_t rop1, 
		twistpoint_fp2_t rop2, 
		const twistpoint_fp2_t op1, 
		const twistpoint_fp2_t op2, 
		const curvepoint_fp_t op3)
{
	// op2 is assumed to be in affine coordinates!

	fp2e_t tmp1, tmp2, tmp3, tmp4, tmp5; // Temporary variables needed for intermediary results

	// Computation of the slope (numerator and denominator), see Guide to ECC, page 91f.
	fp2e_square(tmp2, op1->m_z);
	fp2e_mul(tmp3, tmp2, op1->m_z);

	// Store z_1^3 for use in the evaluation:
	fp2e_set(tmp4, tmp3);	
	
	fp2e_mul(tmp2, tmp2, op2->m_x);
	fp2e_mul(tmp3, tmp3, op2->m_y);
	fp2e_sub(tmp2, tmp2, op1->m_x);
	fp2e_sub(tmp3, tmp3, op1->m_y);
	fp2e_mul(tmp1, tmp2, op1->m_z); // z-coordinate computed, stored in tmp1

	// Compute the linefunction:
	linefunction_evaluate_ate(rop1, tmp3, tmp1, tmp4, op1, op3);

	// Copy z-coordinate to rop:
	fp2e_set(rop2->m_z, tmp1);

	// Continue computation of point coordinates:
	fp2e_square(tmp4, tmp2);
	fp2e_mul(tmp5, tmp4, tmp2);
	fp2e_mul(tmp4, tmp4, op1->m_x);
	fp2e_double(tmp2, tmp4);
	fp2e_square(rop2->m_x, tmp3);
	fp2e_sub(rop2->m_x, rop2->m_x, tmp2);
	fp2e_sub(rop2->m_x, rop2->m_x, tmp5);
	fp2e_sub(tmp4, tmp4, rop2->m_x);
	fp2e_mul(tmp4, tmp4, tmp3);
	fp2e_mul(tmp5, tmp5, op1->m_y);
	fp2e_sub(rop2->m_y, tmp4, tmp5);
}

static void linefunction_double_ate(fp12e_t rop1, twistpoint_fp2_t rop2, const twistpoint_fp2_t op1, const curvepoint_fp_t op3)
{
	fp2e_t tmp1, tmp2, tmp3, tmp4, tmp5; // Temporary variables needed for intermediary results

	// Computation of the slope (numerator and denominator), see DA Peter Schwabe, page 42
	fp2e_square(tmp2, op1->m_y);
	fp2e_mul(tmp3, op1->m_x, tmp2);
	fp2e_double(tmp3, tmp3);
	fp2e_double(tmp3, tmp3);
	
	fp2e_square(tmp4, tmp2);
	fp2e_double(tmp4, tmp4);
	fp2e_double(tmp4, tmp4);
	fp2e_double(tmp4, tmp4);

	fp2e_square(tmp2, op1->m_x);
	fp2e_triple(tmp2, tmp2);

	// Compute z_1^3 for evaluation
	fp2e_square(tmp5, op1->m_z);
	fp2e_mul(tmp5, tmp5, op1->m_z);

	fp2e_mul(tmp1, op1->m_y, op1->m_z);
	fp2e_double(tmp1, tmp1);
	
	// Compute the linefunction:
	linefunction_evaluate_ate(rop1, tmp2, tmp1, tmp5, op1, op3);

	// Copy z-coordinate
	fp2e_set(rop2->m_z, tmp1);

	fp2e_square(rop2->m_x, tmp2);
	fp2e_double(tmp1, tmp3);
	fp2e_sub(rop2->m_x, rop2->m_x, tmp1);
	fp2e_sub(rop2->m_y, tmp3, rop2->m_x);
	fp2e_mul(rop2->m_y, rop2->m_y, tmp2);
	fp2e_sub(rop2->m_y, rop2->m_y, tmp4);
}


void ate(fp12e_t rop, const twistpoint_fp2_t op1, const curvepoint_fp_t op2)
{
	fp12e_setone(rop);

	fp12e_t dummy;

	mpz_t tminus1;
	mpz_init_set(tminus1, trace);
	mpz_sub_ui(tminus1, tminus1, 1);

	twistpoint_fp2_t r;
	twistpoint_fp2_set(r, op1);

	unsigned long int i;

	unsigned long long int start, stop;
	start = cpucycles();
	for(i = mpz_sizeinbase(tminus1, 2) - 1; i > 0; i--)
	{
		linefunction_double_ate(dummy, r, r, op2);
		fp12e_square(rop, rop);
		fp12e_mul(rop, rop, dummy);

		if (mpz_tstbit(tminus1, i - 1)) 
		{
			linefunction_add_ate(dummy, r, r, op1, op2);
			fp12e_mul(rop, rop, dummy);
		}
	}
	stop = cpucycles();
	printf("\nCycles needed: %llu\n\n", stop - start);

	final_expo(rop);
}
